""""""
from abilian.core.singleton import UniqueName


class NS1(UniqueName):
    pass


class NS2(UniqueName):
    pass


def test_singleton() -> None:
    val = NS1("val")
    other_val = NS1("val")
    assert val is other_val
    assert id(val) == id(other_val)


def test_equality() -> None:
    val = NS1("val")
    assert val == "val"
    assert val == "val"


def test_namespaces() -> None:
    ns1_val = NS1("val")
    ns2_val = NS2("val")
    assert ns1_val is not ns2_val
    # equality works because of string compat
    assert ns1_val == ns2_val



""""""
from flask import Flask

from .extension import AttachmentExtension, AttachmentsManager
from .forms import AttachmentForm  # noqa


def register_plugin(app: Flask) -> None:
    AttachmentExtension(app)



""""""
import uuid

from pytest import raises
from sqlalchemy.orm import Session

from . import repository, session_repository
from .service import RepositoryTransaction


def test_transaction_lifetime(session: Session) -> None:
    state = session_repository.app_state
    root_transaction = state.get_transaction(session)
    assert isinstance(root_transaction, RepositoryTransaction)
    assert root_transaction._parent is None

    # create sub-transaction (db savepoint)
    session.begin(nested=True)
    transaction = state.get_transaction(session)
    assert isinstance(transaction, RepositoryTransaction)
    assert transaction._parent is root_transaction

    session.flush()
    transaction = state.get_transaction(session)

    # FIXME
    # assert transaction is root_transaction
    #
    # # create subtransaction (sqlalchemy)
    # session.begin(subtransactions=True)
    # transaction = state.get_transaction(session)
    # assert isinstance(transaction, RepositoryTransaction)
    # assert transaction._parent is root_transaction
    #
    # session.flush()
    # transaction = state.get_transaction(session)
    # assert transaction is root_transaction


def test_accessors_bad_uuid_type(session: Session) -> None:
    uuid_str = b"4f80f02f-52e3-4fe2-b9f2-2c3e99449ce9"

    with raises(ValueError):
        session_repository.get(session, uuid_str)
    with raises(ValueError):
        session_repository.set(session, uuid_str, "")
    with raises(ValueError):
        session_repository.delete(session, uuid_str)


def test_accessors_non_existent_entry(session: Session) -> None:
    # non-existent
    u = uuid.uuid4()
    null = object()
    assert session_repository.get(session, u) is None
    assert session_repository.get(session, u, default=null) is null


def test_accessors_set_get_delete(session: Session) -> None:
    # set
    content = b"my file content"
    u1 = uuid.uuid4()
    session_repository.set(session, u1, content)
    assert session_repository.get(session, u1).open("rb").read() == content
    assert repository.get(u1) is None

    # delete
    session_repository.delete(session, u1)
    assert session_repository.get(session, u1) is None

    u2 = uuid.uuid4()
    repository.set(u2, b"existing content")
    assert session_repository.get(session, u2) is not None

    session_repository.delete(session, u2)
    assert session_repository.get(session, u2) is None
    assert repository.get(u2) is not None


def test_transaction(session: Session) -> None:
    u = uuid.uuid4()
    repository.set(u, b"first draft")
    assert session_repository.get(session, u).open("rb").read() == b"first draft"

    session_repository.set(session, u, b"new content")

    # test nested (savepoint)
    # delete content but rollback transaction
    db_tr = session.begin(nested=True)
    session_repository.delete(session, u)
    assert session_repository.get(session, u) is None

    db_tr.rollback()
    assert session_repository.get(session, u).open("rb").read() == b"new content"

    # delete and commit
    with session.begin(nested=True):
        session_repository.delete(session, u)
        assert session_repository.get(session, u) is None

    assert session_repository.get(session, u) is None
    assert repository.get(u) is not None

    session.commit()
    assert repository.get(u) is None

    # delete: now test subtransactions (sqlalchemy)
    repository.set(u, b"first draft")
    db_tr = session.begin(subtransactions=True)
    session_repository.delete(session, u)
    assert session_repository.get(session, u) is None

    db_tr.rollback()
    assert session_repository.get(session, u).open("rb").read() == b"first draft"

    session.rollback()

    with session.begin(subtransactions=True):
        session_repository.delete(session, u)
        assert session_repository.get(session, u) is None

    assert session_repository.get(session, u) is None
    assert repository.get(u) is not None

    session.commit()
    assert repository.get(u) is None

    # now test 'set'
    session_repository.set(session, u, b"new content")
    session.commit()
    assert repository.get(u) is not None

    # test "set" in two nested transactions. This tests a specific code
    # branch, when a subtransaction overwrite data set in parent
    # transaction
    with session.begin(nested=True):
        session_repository.set(session, u, b"transaction 1")

        with session.begin(nested=True):
            session_repository.set(session, u, b"transaction 2")

        assert session_repository.get(session, u).open("rb").read() == b"transaction 2"


def test_transaction_path(session: Session) -> None:
    """Test RepositoryTransaction create storage only when needed."""
    u = uuid.uuid4()

    state = session_repository.app_state
    root_transaction = state.get_transaction(session)

    # assert not root_transaction.path.exists()

    with session.begin(subtransactions=True):
        transaction = state.get_transaction(session)
        assert not transaction.path.exists()

        session_repository.set(session, u, b"my file content")
        assert transaction.path.exists()

    assert root_transaction.path.exists()

    content = session_repository.get(session, u).open("rb").read()
    assert content == b"my file content"
    assert root_transaction.path.exists()



import json
from datetime import timedelta
from typing import Any, Callable, Dict, Optional

import sqlalchemy as sa

from abilian.core.extensions import db

__all__ = ["Setting", "empty_value"]


class TransformerRegistry:
    def __init__(self) -> None:
        self.encoders: Dict[str, Optional[Callable]] = {}
        self.decoders: Dict[str, Optional[Callable]] = {}

    def encode(self, type_: str, value: Any) -> str:
        return self.encoders.get(type_, str)(value)

    def decode(self, type_: str, value: str) -> Any:
        decoder = self.decoders.get(type_)
        if decoder is not None:
            value = decoder(value)
        return value

    def register(
        self,
        type_: str,
        encoder: Optional[Callable] = None,
        decoder: Optional[Callable] = None,
    ) -> None:
        assert type_
        assert any((encoder, decoder))
        if encoder:
            self.encoders[type_] = encoder
        if decoder:
            self.decoders[type_] = decoder


_transformers = TransformerRegistry()


class _EmptyValue:
    def __bool__(self):
        return False

    def __repr__(self):
        return "<Empty Value>"


#: marker for emptiness, to distinguish from None
empty_value = _EmptyValue()


class Setting(db.Model):
    """A Setting is a very simple key/value object, key being a string
    identifier and the primary key.

    value must be stored as Unicode.
    """

    transformers = _transformers
    key = sa.Column("key", sa.String(length=1000), primary_key=True)

    #: Can be a string (Unicode), int, bool, json... or even a long dotted name
    #: if that's what you need. Type must be set before setting `value`
    _type = sa.Column("type", sa.String(length=1000), nullable=False)

    _value = sa.Column("value", sa.Text())

    @property
    def type(self) -> str:
        return self._type

    @type.setter
    def type(self, type_: str) -> None:
        if not (
            type_ in self.transformers.encoders and type_ in self.transformers.decoders
        ):
            raise ValueError(
                f'Invalid type "{type_}": no encoder and/or decoder registered'
            )
        self._type = type_

    @property
    def value(self):
        if self._value is None:
            return empty_value

        assert isinstance(self._value, str)
        return self.transformers.decode(self.type, self._value)

    @value.setter
    def value(self, value):
        assert self.type
        self._value = self.transformers.encode(self.type, value)
        assert isinstance(self._value, str)


register = _transformers.register


def from_int(i: int) -> str:
    return f"{i}"


register("int", from_int, int)


def from_bool(b: bool) -> str:
    return "true" if b else "false"


def to_bool(s: str) -> bool:
    return s == "true"


register("bool", from_bool, to_bool)


def from_unicode(s: str) -> str:
    return s


def to_unicode(s: str) -> str:
    return s


register("string", from_unicode, to_unicode)


def from_obj(o: Any) -> str:
    return json.dumps(o)


def to_obj(s: str) -> Any:
    return json.loads(s)


register("json", from_obj, to_obj)


def from_timedelta(s: timedelta) -> str:
    return json.dumps({"days": s.days, "seconds": s.seconds})


def to_timedelta(s: str):
    return timedelta(**json.loads(s))


register("timedelta", from_timedelta, to_timedelta)
